@@ -512,10 +512,12 @@ void test_mul(void)
 {\
     int res, val, resz;\
     val = op0;\
-    asm("xorl %1, %1 ; " #op " %" size "2, %" size "0 ; setz %b1" \
+    asm("xorl %1, %1\n"\
+        "movl $0x12345678, %0\n"\
+        #op " %" size "2, %" size "0 ; setz %b1" \
         : "=r" (res), "=q" (resz)\
         : "g" (val));\
-    printf("%-10s A=%08x R=%08x %d\n", #op, val, resz ? 0 : res, resz);\
+    printf("%-10s A=%08x R=%08x %d\n", #op, val, res, resz);\
 }
 
 void test_bsx(void)
@@ -686,6 +688,14 @@ void test_fenv(void)
     TEST_ENV(&float_env16, "data16 fnsave", "data16 frstor");
     TEST_ENV(&float_env32, "fnstenv", "fldenv");
     TEST_ENV(&float_env32, "fnsave", "frstor");
+
+    /* test for ffree */
+    for(i=0;i<5;i++)
+        asm volatile ("fldl %0" : : "m" (dtab[i]));
+    asm volatile("ffree %st(2)");
+    asm volatile ("fnstenv %0\n" : : "m" (float_env32));
+    asm volatile ("fninit");
+    printf("fptag=%04x\n", float_env32.fptag);
 }
 
 
